import math

keylane = {0:"16", 1:"11", 2:"12", 3:"13", 4:"14", 5:"15", 6:"18", 7:"19"}

class Pattern:

    #objects = [ [ ["00","00", ... *res*4],["00","00", ... *res], ... *8(7+S) ],
    #            [ ["00","00", ... *res*4],["00","00", ... *res], ... *8(7+S) ], ... *song_length ]
    #ext_ch = [["etc command"], [...](for i in range(song_length))]


    def __init__(self, resolution, song_length, bar_length=1):

        assert (resolution*bar_length) % 1 == 0, "Invalid BarLength | Res:{}, bar_length:{}".format(resolution, bar_length)

        self.song_length = song_length
        self.bar_length = bar_length
        self.resolution = self.num_in_bar(resolution) # if res=192 & bar length = 0.75, self.res = 144
        self.num_quarter = self.num_in_bar(4)

        if bar_length != 1:
            self.ext_ch = [["#{:03d}02:{}".format(i,bar_length)] for i in range(song_length+2)] # +2 = 000, last bar
        else:
            self.ext_ch = [[] for i in range(song_length+2)]

        self.objects = [[["00" for i in range(self.resolution)] for j in range(8)] 
                            for k in range(song_length+2)]

    
    def num_in_bar(self, count_resolution:int, bar_length=0) -> int:
        '''
        return how many note with count resolution is in the bar.
        '''

        if bar_length == 0:
            bar_length = self.bar_length
        
        return math.ceil(count_resolution * bar_length) if math.ceil(count_resolution * bar_length) >= 1 else 1


    def num_in_quarter(self, count_resolution:int) -> int:
        '''
        return how many note with count resolution is in the 1 quarter note.
        '''
        
        return int(self.num_in_bar(count_resolution) / self.num_quarter)


    def count_on_bar(self, countWithCountResolution:int, countResolution:int, baseResolution:int=0) -> int:
        '''
        return count with whole resolution from specific resolution & count.

        *You don't have to care about countRes|baseRes.
        It's automatically adjusted with bar_length.
        '''

        if baseResolution == 0:
            baseResolution = self.resolution
        else:
            baseResolution = self.num_in_bar(baseResolution)

        #assert self.num_in_bar(baseResolution,1) % countResolution == 0, "Excessive Count Resolution | res:{} cnt_res:{}".format(self.num_in_bar(baseResolution,1), countResolution)
        assert self.num_in_bar(countResolution) > countWithCountResolution, "Count {} out of range: {} (count must be in 0 ~ res-1)".format(countWithCountResolution, countResolution)

        return int(baseResolution / self.num_in_bar(countResolution)) * countWithCountResolution

    
    def location(self, countWithCountResolution: int, countResolution: int, baseResolution:int=0) -> int:
        '''
        alias of Function "count_on_bar".
        '''

        return self.count_on_bar(countWithCountResolution, countResolution, baseResolution)


    def note(self, lane: int, bar: int, count: int, count_res: int, wav: str, force=False) -> bool:
        '''
        write note object.
        if note already exists, return False.
        if force Flag is enabled, It will override note if exists.

        *count must start from 0.
        '''

        if (force == False) and self.objects[bar][lane][self.location(count, count_res)] != "00":
            return False
        
        self.objects[bar][lane][self.location(count, count_res)] = wav
        return True


    def bgm(self, bar: int, count: int, count_res: int, wav: str):
        '''
        write bgm object. 
        
        *count must start from 0.
        '''
        _location = self.location(count, count_res, count_res)
        self.objects[bar].append([wav if (i == _location) else "00" for i in range(self.num_in_bar(count_res))])
    

    def ext_command(self, command: str, bar: int, count:int = 0, count_res:int = 1):
        self.ext_ch[bar].append("#{:03d}{}".format(bar, command))


    def export(self):
        data = []
        for bar,lanes in enumerate(self.objects):

            data += [i+"\r\n" for i in self.ext_ch[bar]]

            for key, parameter in enumerate(lanes):
                _lane = keylane[key] if key <= 7 else "01"
                command = "#{:03d}{}:{}\r\n".format(bar, _lane, "".join(parameter))
                data.append(command)
        
            data.append("\r\n")
        
        return data